#ifndef __NLL_LABEL_H_INCLUDED__
#define __NLL_LABEL_H_INCLUDED__

struct label {
  struct label *next;
  struct label *prev;
  struct spot spot;
  int refcount;
  char name[LABEL_NAMELEN + 1];
};

int label_check(void);
int label_init(void);
int label_done(void);
int label_get(const char *name, label_t *labelp);
int label_set(const char *name, spot_t spot, label_t *labelp);
int label_clear(label_t label);
int label_clean(label_t label);
int label_del(label_t label);
int label_dump(FILE *fp, label_t label);

#endif
